/* scherzo - Music notation training
 *
 * Copyright © 2010 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <asoundlib.h>

#include "score.h"
#include "mnemon.h"

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof(arr[0]))

#define unused(foo) foo __attribute__((unused))

#define MIDI_BUF_SIZE 4096

typedef struct challenge
{
    bin_t *bin;
    int item_index;
    score_staff_t *staff;
    score_note_t *note;

    int satisfied;
    int mistaken;
} challenge_t;

typedef struct note_group
{
    void *ctx;
    score_note_t **notes;
    int size;
    int num_notes;
} note_group_t;

typedef struct scherzo
{
    void *ctx;

    GtkWidget *window;
    score_t *score;
    int staff_height;
    score_staff_t *treble;
    score_staff_t *bass;
    score_chord_t *chord;

    /* The word "keyboard" here is referring to a "computer
     * keyboard". Any "piano keyboard" key knows its own octave and
     * accidental already. */
    int keyboard_octave;
    score_pitch_accidental_t keyboard_accidental;

    int midi_fd;
    snd_midi_event_t *snd_midi_event;

    mnemon_t mnemon;
    challenge_t challenge;

    note_group_t notes_pressed;
    note_group_t notes_pedaled;

    int pedal_pressed;
} scherzo_t;

/* Forward declarations. */
static score_note_t *
scherzo_press_note (scherzo_t *scherzo, score_pitch_t pitch, int octave);

static void
scherzo_release_note (scherzo_t *scherzo, score_pitch_t pitch, int octave);

static void
scherzo_press_pedal (scherzo_t *scherzo);

static void
scherzo_release_pedal (scherzo_t *scherzo);

static void
_judge_note (scherzo_t *scherzo, score_note_t *note);

static void
_score_challenge (scherzo_t *scherzo);

static int
on_delete_event_quit (unused (GtkWidget *widget),
		      unused (GdkEvent *event),
		      unused (gpointer user_data))
{
    gtk_main_quit ();

    /* Returning FALSE allows the default handler for delete-event
     * to proceed to cleanup the widget. */
    return FALSE;
}

static int
on_expose_event_draw (GtkWidget *widget,
		      unused (GdkEventExpose *expose),
		      void * user_data)
{
    scherzo_t *scherzo = user_data;
    score_t *score = scherzo->score;
    cairo_t *cr;
    GtkAllocation allocation;
    static const int pad = 10;
    int widget_width;

    gtk_widget_get_allocation (widget, &allocation);
    widget_width = allocation.width;

    cr = gdk_cairo_create (widget->window);

    /* White background */
    cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
    cairo_paint (cr);

    /* Add some padding on the sides and top */
    cairo_translate (cr, pad, scherzo->staff_height);
    score_set_staff_height (score, scherzo->staff_height);
    score_set_width (score, widget_width - 2 * pad);

    score_draw (score, cr);
 
    return TRUE;
}

static int
on_key_press_event (GtkWidget *widget,
		    GdkEventKey *key,
		    void *user_data)
{
    scherzo_t *scherzo = user_data;
    int octave;
    /* Initialize to keep the compiler quiet. */
    score_pitch_name_t pitch_name = SCORE_PITCH_C;
    score_pitch_t pitch;

    if (scherzo->challenge.note)
	octave = scherzo->challenge.note->octave;
    else
	octave = scherzo->keyboard_octave;

    switch (key->keyval) {
    case GDK_KEY_plus:
    case GDK_KEY_KP_Add:
    case GDK_KEY_equal:
    case GDK_KEY_KP_Equal:
	scherzo->staff_height += 4;
	gtk_widget_queue_draw (widget);
	return TRUE;
	break;
    case GDK_KEY_minus:
    case GDK_KEY_KP_Subtract:
	scherzo->staff_height -= 4;
	gtk_widget_queue_draw (widget);
	return TRUE;
	break;
    case GDK_KEY_q:
    case GDK_KEY_Q:
    case GDK_KEY_Escape:
	gtk_main_quit ();
	return FALSE;
    case GDK_KEY_c:
    case GDK_KEY_C:
	pitch_name = SCORE_PITCH_NAME_C;
	break;
    case GDK_KEY_d:
    case GDK_KEY_D:
	pitch_name = SCORE_PITCH_NAME_D;
	break;
    case GDK_KEY_e:
    case GDK_KEY_E:
	pitch_name = SCORE_PITCH_NAME_E;
	break;
    case GDK_KEY_f:
    case GDK_KEY_F:
	pitch_name = SCORE_PITCH_NAME_F;
	break;
    case GDK_KEY_g:
    case GDK_KEY_G:
	pitch_name = SCORE_PITCH_NAME_G;
	break;
    case GDK_KEY_a:
    case GDK_KEY_A:
	pitch_name = SCORE_PITCH_NAME_A;
	break;
    case GDK_KEY_b:
    case GDK_KEY_B:
	pitch_name = SCORE_PITCH_NAME_B;
	break;
    case GDK_KEY_0:
    case GDK_KEY_1:
    case GDK_KEY_2:
    case GDK_KEY_3:
    case GDK_KEY_4:
    case GDK_KEY_5:
    case GDK_KEY_6:
    case GDK_KEY_7:
    case GDK_KEY_8:
	scherzo->keyboard_octave = key->keyval - GDK_KEY_0;
	break;
    case GDK_KEY_space:
	scherzo_press_pedal (scherzo);
	break;
    case GDK_KEY_Up:
	if (scherzo->keyboard_accidental < SCORE_PITCH_ACCIDENTAL_DOUBLE_SHARP)
	    scherzo->keyboard_accidental++;
	break;
    case GDK_KEY_Down:
	if (scherzo->keyboard_accidental > SCORE_PITCH_ACCIDENTAL_DOUBLE_FLAT)
	    scherzo->keyboard_accidental--;
	break;
    }

    pitch = SCORE_PITCH (pitch_name, scherzo->keyboard_accidental);

    if ((key->keyval >= GDK_KEY_A && key->keyval <= GDK_KEY_G) ||
	(key->keyval >= GDK_KEY_a && key->keyval <= GDK_KEY_g))
    {
	score_note_t *note;

	note = scherzo_press_note (scherzo, pitch, octave);
	_judge_note (scherzo, note);
	gtk_widget_queue_draw (scherzo->window);

	return TRUE;
    }


    /* Allow an unhandled event to propagate to other handlers. */
    return FALSE;
}

static int
on_key_release_event (unused (GtkWidget *widget),
		      GdkEventKey *key,
		      void *user_data)
{
    scherzo_t *scherzo = user_data;
    int octave;
    /* Initialize to keep the compiler quiet. */
    score_pitch_name_t pitch_name = SCORE_PITCH_NAME_C;
    score_pitch_t pitch;

    if (scherzo->challenge.note)
	octave = scherzo->challenge.note->octave;
    else
	octave = scherzo->keyboard_octave;

    switch (key->keyval) {
    case GDK_KEY_c:
    case GDK_KEY_C:
	pitch_name = SCORE_PITCH_NAME_C;
	break;
    case GDK_KEY_d:
    case GDK_KEY_D:
	pitch_name = SCORE_PITCH_NAME_D;
	break;
    case GDK_KEY_e:
    case GDK_KEY_E:
	pitch_name = SCORE_PITCH_NAME_E;
	break;
    case GDK_KEY_f:
    case GDK_KEY_F:
	pitch_name = SCORE_PITCH_NAME_F;
	break;
    case GDK_KEY_g:
    case GDK_KEY_G:
	pitch_name = SCORE_PITCH_NAME_G;
	break;
    case GDK_KEY_a:
    case GDK_KEY_A:
	pitch_name = SCORE_PITCH_NAME_A;
	break;
    case GDK_KEY_b:
    case GDK_KEY_B:
	pitch_name = SCORE_PITCH_NAME_B;
	break;
    case GDK_KEY_space:
	scherzo_release_pedal (scherzo);
	break;
    }

    pitch = SCORE_PITCH (pitch_name, scherzo->keyboard_accidental);

    if ((key->keyval >= GDK_KEY_A && key->keyval <= GDK_KEY_G) ||
	(key->keyval >= GDK_KEY_a && key->keyval <= GDK_KEY_g))
    {
	scherzo_release_note (scherzo, pitch, octave);
	_score_challenge (scherzo);
	gtk_widget_queue_draw (scherzo->window);

	return TRUE;
    }


    /* Allow an unhandled event to propagate to other handlers. */
    return FALSE;
}

static unsigned char
_score_pitch_and_octave_to_midi (score_pitch_t pitch,
				 int octave)
{
    unsigned char midi_note = 12 * (octave + 1);

    switch (SCORE_PITCH_NAME (pitch)) {
    case SCORE_PITCH_NAME_C:
	break;
    case SCORE_PITCH_NAME_D:
	midi_note += 2;
	break;
    case SCORE_PITCH_NAME_E:
	midi_note += 4;
	break;
    case SCORE_PITCH_NAME_F:
	midi_note += 5;
	break;
    case SCORE_PITCH_NAME_G:
	midi_note += 7;
	break;
    case SCORE_PITCH_NAME_A:
	midi_note += 9;
	break;
    case SCORE_PITCH_NAME_B:
	midi_note += 11;
	break;
    }

    switch (SCORE_PITCH_ACCIDENTAL (pitch)) {
    case SCORE_PITCH_ACCIDENTAL_DOUBLE_FLAT:
	midi_note -= 2;
	break;
    case SCORE_PITCH_ACCIDENTAL_FLAT:
	midi_note -= 1;
	break;
    case SCORE_PITCH_ACCIDENTAL_NATURAL:
	break;
    case SCORE_PITCH_ACCIDENTAL_SHARP:
	midi_note += 1;
	break;
    case SCORE_PITCH_ACCIDENTAL_DOUBLE_SHARP:
	midi_note += 2;
	break;
    }

    return midi_note;
}

static void
_midi_to_score_pitch_and_octave (unsigned char midi_note,
				 score_pitch_t *pitch,
				 int *octave)
{
    *octave = midi_note / 12 - 1;

    switch (midi_note % 12)
    {
    case 0:
	*pitch = SCORE_PITCH_C;
	break;
    case 1:
	*pitch = SCORE_PITCH_Cs;
	break;
    case 2:
	*pitch = SCORE_PITCH_D;
	break;
    case 3:
	*pitch = SCORE_PITCH_Ds;
	break;
    case 4:
	*pitch = SCORE_PITCH_E;
	break;
    case 5:
	*pitch = SCORE_PITCH_F;
	break;
    case 6:
	*pitch = SCORE_PITCH_Fs;
	break;
    case 7:
	*pitch = SCORE_PITCH_G;
	break;
    case 8:
	*pitch = SCORE_PITCH_Gs;
	break;
    case 9:
	*pitch = SCORE_PITCH_A;
	break;
    case 10:
	*pitch = SCORE_PITCH_As;
	break;
    case 11:
	*pitch = SCORE_PITCH_B;
	break;
    }
}

/* Determine a chord name (if any) from the current notes pressed */

typedef struct analyzed_note {
    /* Original note being analzyed. */
    score_note_t *note;

    /* Absolute pitch (expressed as midi number). */
    int midi_pitch;

    /* Pitch relative to bass note. */
    int relative_pitch;
} analyzed_note_t;

static int
_compare_analyzed_note_by_midi_pitch (const void *va, const void *vb)
{
    const analyzed_note_t *a = va, *b = vb;

    return a->midi_pitch - b->midi_pitch;
}

static int
_compare_analyzed_note_by_relative_pitch (const void *va, const void *vb)
{
    const analyzed_note_t *a = va, *b = vb;

    return a->relative_pitch - b->relative_pitch;
}

typedef struct modified_degree
{
    int degree;
    int modification;
} modified_degree_t;

static int
_modified_degree_to_half_steps (const modified_degree_t *degree)
{
    int half_steps;

    /* Number of half steps from root to specified degree within a
     * diatonic scaled. */
    switch (degree->degree) {
    case 1:
	half_steps = 0;
	break;
    case 2:
	half_steps = 2;
	break;
    case 3:
	half_steps = 4;
	break;
    case 4:
	half_steps = 5;
	break;
    case 5:
	half_steps = 7;
	break;
    case 6:
	half_steps = 9;
	break;
    case 7:
	half_steps = 11;
	break;
    default:
	fprintf (stderr, "Internal: Invalid degree %d\n", degree->degree);
	exit (1);
	break;
    }

    return half_steps + degree->modification;
}

static int
_chord_signature_matches (analyzed_note_t *notes,
			  int num_notes,
			  modified_degree_t *degrees,
			  int num_degrees,
			  int *inversion_ret)
{
#define MAX_DEGREES 4
    int relative_pitches[MAX_DEGREES];
    int inversion, max_inversions;
    int i;

    assert (num_degrees <= MAX_DEGREES);

    if (num_notes != num_degrees)
	    return 0;

    max_inversions = num_degrees;

    /* We never spell simple intervals as inversions. */
    if (num_degrees == 2)
	max_inversions = 1;

    for (inversion = 0; inversion < max_inversions; inversion++) {
	for (i = 0; i < num_degrees; i++) {
	    /* The num_degrees is in the addition just to ensure all
	     * inputs to the modulus operator remain positive. */
	    int index = (i + num_degrees - inversion) % num_degrees;

	    /* Again, adding a 12 to keep things positive. */
	    relative_pitches[index] =
		(12 +
		 _modified_degree_to_half_steps (&degrees[i]) -
		 _modified_degree_to_half_steps (&degrees[inversion])) % 12;
		
	}

	for (i = 0; i < num_notes; i++)
	    if (notes[i].relative_pitch != relative_pitches[i])
		goto NEXT_INVERSION;

	*inversion_ret = inversion;

	return 1;

    NEXT_INVERSION:
	;
    }

    return 0;
}

static void
scherzo_analyze_chord (scherzo_t *scherzo)
{
    void *local = talloc_new (NULL);
    analyzed_note_t *notes;
    note_group_t *note_group;
    unsigned i, j, num_notes;
    int bass_pitch, inversion;
    const char *chord_name = NULL;

    if (scherzo->pedal_pressed)
	note_group = &scherzo->notes_pedaled;
    else
	note_group = &scherzo->notes_pressed;

    num_notes = note_group->num_notes;

    struct { modified_degree_t degrees[1]; const char *name; } octaves[] = {
	{ {{1, 0}}, "Octave"}
    };

    struct { modified_degree_t degrees[2]; const char *name; } intervals[] = {
	{ {{1, 0}, {2, -1}}, "Minor 2nd"},
	{ {{1, 0}, {2,  0}}, "Major 2nd"},
	{ {{1, 0}, {3, -1}}, "Minor 3rd"},
	{ {{1, 0}, {3,  0}}, "Major 3rd"},
	{ {{1, 0}, {4,  0}}, "Perfect 4th"},
	{ {{1, 0}, {5, -1}}, "Diminished 5th"},
	{ {{1, 0}, {5,  0}}, "Perfect 5th"},
	{ {{1, 0}, {6, -1}}, "Minor 6th"},
	{ {{1, 0}, {6,  0}}, "Major 6th"},
	{ {{1, 0}, {7, -1}}, "Minor 7th"},
	{ {{1, 0}, {7,  0}}, "Major 7th"}
    };

    struct { modified_degree_t degrees[3]; const char *name; } triads[] = {
	{ {{1, 0}, {3,  0}, {5, +1}}, "Augmented triad" },
	{ {{1, 0}, {3,  0}, {5,  0}}, "Major triad" },
	{ {{1, 0}, {3, -1}, {5,  0}}, "Minor triad" },
	{ {{1, 0}, {3, -1}, {5, -1}}, "Diminished triad" }
    };

    struct { modified_degree_t degrees[4]; const char *name; } sevenths[] = {
	{ {{1, 0}, {3,  0}, {5, +1}, {7,  0}}, "Augmented/major 7" },
	{ {{1, 0}, {3,  0}, {5, +1}, {7, -1}}, "Augmented 7" },
	{ {{1, 0}, {3,  0}, {5,  0}, {7,  0}}, "Major 7" },
	{ {{1, 0}, {3,  0}, {5,  0}, {7, -1}}, "Dominant 7" },
	{ {{1, 0}, {3, -1}, {5,  0}, {7,  0}}, "Minor/major 7" },
	{ {{1, 0}, {3, -1}, {5,  0}, {7, -1}}, "Minor 7" },
	{ {{1, 0}, {3, -1}, {5, -1}, {7,  0}}, "Diminished/major 7" },
	{ {{1, 0}, {3, -1}, {5, -1}, {7, -1}}, "Half-diminished 7" },
	{ {{1, 0}, {3, -1}, {5, -1}, {7, -2}}, "Diminished 7" }
    };

    if (scherzo->chord) {
	score_remove_chord (scherzo->chord);
	scherzo->chord = NULL;
    }

    if (num_notes <= 1)
	goto DONE;

    notes = talloc_array (local, analyzed_note_t, num_notes);
    if (notes == NULL)
	goto DONE;

    for (i = 0; i < num_notes; i++) {
	score_note_t *note = note_group->notes[i];
	notes[i].note = note;
	notes[i].midi_pitch = _score_pitch_and_octave_to_midi (note->pitch,
							       note->octave);
	/* Relative pitch will be filled in after sorting. */
	notes[i].relative_pitch = 0;
    }

    /* First, sort by midi pitch to find the bass note. */
    qsort (notes, num_notes, sizeof (analyzed_note_t),
	   _compare_analyzed_note_by_midi_pitch);
    
    bass_pitch = notes[0].midi_pitch;

    /* With the bass note identified, we can find all relative pitches. */
    for (i = 0; i < num_notes; i++) {
	notes[i].relative_pitch = notes[i].midi_pitch - bass_pitch;
	while (notes[i].relative_pitch >= 12)
	    notes[i].relative_pitch -= 12;
    }

    /* Now, sort again by relative pitch. */
    qsort (notes, num_notes, sizeof (analyzed_note_t),
	   _compare_analyzed_note_by_relative_pitch);

    /* Finally, eliminate all duplicate notes. */
    for (i = 0; i < num_notes - 1; i++) {
	    if (notes[i+1].relative_pitch == notes[i].relative_pitch) {
		    j = i+1;
		    while (j < num_notes &&
			   notes[j].relative_pitch == notes[i].relative_pitch)
		    {
			    j++;
		    }
		    /* The loop incremented j one past the last
		     * duplicate. Decrement so that it points to the
		     * last duplicate (and is guaranteed to not exceed
		     * the array bounds).*/
		    j--;

		    if (j < num_notes - 1) {
			    memmove (&notes[i+1], &notes[j+1],
				     (num_notes - j) * sizeof (analyzed_note_t));
		    }

		    num_notes -= (j - i);
	    }
    }

    switch (num_notes) {
    case 1:
	for (i = 0; i < ARRAY_SIZE (octaves); i++) {
	    if (_chord_signature_matches (notes, num_notes,
					  octaves[i].degrees, 1, &inversion))
	    {
		chord_name = octaves[i].name;
		break;
	    }
	}
	break;
    case 2:
	for (i = 0; i < ARRAY_SIZE (intervals); i++) {
	    if (_chord_signature_matches (notes, num_notes,
					  intervals[i].degrees, 2, &inversion))
	    {
		chord_name = intervals[i].name;
		break;
	    }
	}
	break;
    case 3:
	for (i = 0; i < ARRAY_SIZE (triads); i++) {
	    if (_chord_signature_matches (notes, num_notes,
					  triads[i].degrees, 3, &inversion))
	    {
		chord_name = triads[i].name;
		break;
	    }
	}
	break;
    case 4:
	for (i = 0; i < ARRAY_SIZE(sevenths); i++) {
	    if (_chord_signature_matches (notes, num_notes,
					  sevenths[i].degrees, 4, &inversion))
	    {
		chord_name = sevenths[i].name;
		break;
	    }
	}
	break;
    }

    if (chord_name) {
	if (inversion) {
	    const char *inversion_str;
	    switch (inversion) {
	    case 1:
		inversion_str = "1st inversion";
		break;
	    case 2:
		inversion_str = "2nd inversion";
		break;
	    case 3:
		inversion_str = "3rd inversion";
		break;
	    default:
		fprintf (stderr, "Internal error: Unexpected inversion: %d\n",
			 inversion);
		exit(1);
	    }
	    chord_name = talloc_asprintf (local, "%s %s",
					  chord_name, inversion_str);
	} else {
	    chord_name = talloc_strdup (local, chord_name);
	}
    } else {
	chord_name = talloc_strdup (local, "Unknown chord");
    }

    scherzo->chord = score_add_chord (scherzo->treble, chord_name);

DONE:
    talloc_free (local);
}

static void
note_group_init (void *ctx, note_group_t *group)
{
    group->ctx = ctx;
    group->notes = NULL;
    group->size = 0;
    group->num_notes = 0;
}

static void
note_group_add_note (note_group_t *group, score_note_t *note)
{
    int i;

    /* Do nothing if note is already in group. */
    for (i = 0; i < group->num_notes; i++) {
	if (group->notes[i]->pitch == note->pitch &&
	    group->notes[i]->octave == note->octave)
	{
	    return;
	}
    }

    group->num_notes++;

    if (group->num_notes > group->size) {
	group->size++;
	group->notes = talloc_realloc (group->ctx, group->notes,
				       score_note_t*, group->size);

	if (group->notes == NULL) {
	    fprintf (stderr, "Out of memory.\n");
	    exit (1);
	}
    }

    group->notes[group->num_notes - 1] = note;
}

static void
note_group_remove_note_at (note_group_t *group, int i)
{
    if (i >= group->num_notes) {
	fprintf (stderr, "Internal error: No note to remove at index %d\n", i);
	exit (1);
    }

    if (i < group->num_notes - 1) {
	memmove (group->notes + i, group->notes + i + 1,
		 (group->num_notes - 1 - i) * sizeof (score_note_t*));
    }
    group->num_notes--;
}

static score_note_t *
scherzo_press_note (scherzo_t *scherzo, score_pitch_t pitch, int octave)
{
    score_staff_t *staff;
    score_note_t *note;
    int i;

    if (scherzo->challenge.note) {
	staff = scherzo->challenge.staff;
    } else if (octave >= 4) {
	staff = scherzo->treble;
    } else {
	staff = scherzo->bass;
    }

    /* Do nothing if this note is already pressed. */
    for (i = 0; i < scherzo->notes_pressed.num_notes; i++) {
	if (scherzo->notes_pressed.notes[i]->pitch == pitch &&
	    scherzo->notes_pressed.notes[i]->octave == octave)
	{
	    return scherzo->notes_pressed.notes[i];
	}
    }

    note = score_add_note (staff, pitch, octave, SCORE_DURATION_WHOLE);

    note_group_add_note (&scherzo->notes_pressed, note);

    if (scherzo->pedal_pressed)
	note_group_add_note (&scherzo->notes_pedaled, note);

    scherzo_analyze_chord (scherzo);

    return note;
}

static void
scherzo_release_note (scherzo_t *scherzo, score_pitch_t pitch, int octave)
{
    score_note_t *note;
    int i;
    int found = 0;

    for (i = scherzo->notes_pressed.num_notes - 1; i >=0; i--) {
	note = scherzo->notes_pressed.notes[i];
	if (note->pitch == pitch && note->octave == octave) {
	    found = 1;
	    if (! scherzo->pedal_pressed)
		score_remove_note (note);
	    note_group_remove_note_at (&scherzo->notes_pressed, i);
	}
    }

    if (found == 0) {
	fprintf (stderr, "Internal error: Failed to find note to release.\n");
    }

    scherzo_analyze_chord (scherzo);
}

static score_note_t *
scherzo_press_note_midi (scherzo_t *scherzo, unsigned char midi_note)
{
    score_pitch_t pitch;
    int octave;

    _midi_to_score_pitch_and_octave (midi_note, &pitch, &octave);

    return scherzo_press_note (scherzo, pitch, octave);
}

static void
scherzo_release_note_midi (scherzo_t *scherzo, unsigned char midi_note)
{
    score_pitch_t pitch;
    int octave;
 
    _midi_to_score_pitch_and_octave (midi_note, &pitch, &octave);

    scherzo_release_note (scherzo, pitch, octave);
}

static void
scherzo_press_pedal (scherzo_t *scherzo)
{
    int i;

    scherzo->pedal_pressed = 1;

    /* Copy all pressed notes to pedaled notes */
    for (i = 0; i < scherzo->notes_pressed.num_notes; i++)
	note_group_add_note (&scherzo->notes_pedaled, scherzo->notes_pressed.notes[i]);
}

static void
scherzo_release_pedal (scherzo_t *scherzo)
{
    score_note_t *note, *new_note;
    int i;

    /* Make new notes in score for all pressed notes. */
    for (i = 0; i < scherzo->notes_pressed.num_notes; i++) {
	note = scherzo->notes_pressed.notes[i];
	new_note = score_add_note (note->staff, note->pitch, note->octave, note->duration);
	scherzo->notes_pressed.notes[i] = new_note;
    }

    /* Then remove all previously pedaled notes from the score. */
    for (i = scherzo->notes_pedaled.num_notes - 1; i >=0; i--) {
	note = scherzo->notes_pedaled.notes[i];
	score_remove_note (note);
	note_group_remove_note_at (&scherzo->notes_pedaled, i);
    }

    scherzo->pedal_pressed = 0;

    scherzo_analyze_chord (scherzo);

    gtk_widget_queue_draw (scherzo->window);
}

void
_select_challenge (scherzo_t *scherzo)
{
    category_t *category_unused;
    bool_t introduced_unused;
    item_t *item;
    challenge_t *challenge = &scherzo->challenge;
    score_pitch_t pitch;
    int octave;
    char *s;

    if (challenge->note) {
	score_remove_note (challenge->note);
	challenge->note = NULL;
    }

    mnemon_select_item (&scherzo->mnemon,
			&challenge->bin,
			&challenge->item_index,
			&category_unused,
			&introduced_unused);

    item = challenge->bin->items[challenge->item_index];

    s = item->challenge;
    if (strncmp (s, "treble:", 7) == 0) {
	s += 7;
	challenge->staff = scherzo->treble;
    } else if (strncmp (s, "bass:", 5) == 0) {
	s += 5;
	challenge->staff = scherzo->bass;
    } else {
	fprintf (stderr,
		 "Malformed staff name: %s (expected 'treble:' or 'bass:')\n",
		 s);
	exit (1);
    }

    switch (*s) {
    case 'C':
	pitch = SCORE_PITCH_C;
	break;
    case 'D':
	pitch = SCORE_PITCH_D;
	break;
    case 'E':
	pitch = SCORE_PITCH_E;
	break;
    case 'F':
	pitch = SCORE_PITCH_F;
	break;
    case 'G':
	pitch = SCORE_PITCH_G;
	break;
    case 'A':
	pitch = SCORE_PITCH_A;
	break;
    case 'B':
	pitch = SCORE_PITCH_B;
	break;
    default:
	fprintf (stderr, "Malformed pitch name: %s (expected 'A' - 'G')\n", s);
	exit (1);
    }
    s++;

    if (*s < '0' || *s > '9') {
	fprintf (stderr, "Malformed octave number: %s (expected '0' - '9')\n", s);
	exit (1);
    }

    octave = *s - '0';

    challenge->note = score_add_note (challenge->staff, pitch, octave,
				      SCORE_DURATION_WHOLE);
    challenge->satisfied = 0;
    challenge->mistaken = 0;
}

/* Determine whether the user hit the correct note. */
static void
_judge_note (scherzo_t *scherzo, score_note_t *note)
{
    challenge_t *challenge = &scherzo->challenge;

    if (! scherzo->challenge.note) {
	score_set_note_color_rgb (note, 0.0, 0.0, 0.0); /* black */
	return;
    }

    if (note->pitch == challenge->note->pitch &&
	note->octave == challenge->note->octave)
    {
	challenge->satisfied = 1;
	score_set_note_color_rgb (note, 18/256., 130/256., 28/256.); /* green */
    }
    else
    {
	challenge->mistaken = 1;
	score_set_note_color_rgb (note, 184/256., 4/256., 22/256.); /* red */
    }
}

/* If the user got the right note (eventually), then score it in
 * mnemon and show the next note. */
static void
_score_challenge (scherzo_t *scherzo)
{
    challenge_t *challenge = &scherzo->challenge;

    if (! challenge->note)
	return;

    if (! challenge->satisfied)
	return;

    mnemon_score_item (&scherzo->mnemon, challenge->bin, challenge->item_index,
		       ! challenge->mistaken);

    _select_challenge (scherzo);
}

static int
on_midi_input (unused (GIOChannel *channel),
	       unused (GIOCondition condition),
	       void *user_data)
{
    unsigned char buf[MIDI_BUF_SIZE], *next;
    scherzo_t *scherzo = user_data;
    ssize_t remaining;
    snd_seq_event_t event;
    score_note_t *note;
    int need_redraw = FALSE;

    remaining = read (scherzo->midi_fd, buf, MIDI_BUF_SIZE);

    next = buf;
    while (remaining) {
	long consumed;

	consumed = snd_midi_event_encode (scherzo->snd_midi_event,
					  next, remaining, &event);

	remaining -= consumed;
	next += consumed;

	switch (event.type) {
	case SND_SEQ_EVENT_NONE:
	    /* Incomplete event. Nothing to do. */
	    break;
	case SND_SEQ_EVENT_NOTEON:
	    note = scherzo_press_note_midi (scherzo, event.data.note.note);
	    _judge_note (scherzo, note);
	    need_redraw = TRUE;
	    break;
	case SND_SEQ_EVENT_NOTEOFF:
	    scherzo_release_note_midi (scherzo, event.data.note.note);
	    _score_challenge (scherzo);
	    need_redraw = TRUE;
	    break;
	case SND_SEQ_EVENT_CLOCK:
	    /* Ignore for now as my piano sends a constant stream of these. */
	    break;
	case SND_SEQ_EVENT_SENSING:
	    /* Ignore for now as my piano sends a constant stream of these. */
	    break;
	case SND_SEQ_EVENT_CONTROLLER:
	    /* XXX: My piano gives me 64 for the sustain pedal, is
	     * that universal? */
	    if (event.data.control.param == 64) {
		if (event.data.control.value == 0)
		    scherzo_release_pedal (scherzo);
		else
		    scherzo_press_pedal (scherzo);
	    } else {
		fprintf (stderr, "Fixme: Unhandled MIDI Control event (param=%d, value=%d)\n",
			 event.data.control.param, event.data.control.value);
	    }
	    break;
	default:
	    fprintf (stderr, "Fixme: Do not yet know how to handle MIDI event %d\n",
		     event.type);
	    break;
	}
    }

    if (need_redraw)
	gtk_widget_queue_draw (scherzo->window);

    /* Return TRUE to continue to get called in the future. */
    return TRUE;
}

int
main (int argc, char *argv[])
{
    GtkWidget *drawing_area;
    scherzo_t scherzo;
    int err;

    srand (time (NULL));

    gtk_init (&argc, &argv);

    scherzo.ctx = talloc_new (NULL);

    scherzo.score = score_create (scherzo.ctx);
    scherzo.staff_height = 100;
    score_set_staff_height (scherzo.score, scherzo.staff_height);

    score_add_brace (scherzo.score, 2);
    scherzo.treble = score_add_staff (scherzo.score, SCORE_CLEF_G);
    scherzo.bass = score_add_staff (scherzo.score, SCORE_CLEF_F);

    scherzo.chord = NULL;

    /* Default to octave 4 and natural for computer keyboard keypresses. */
    scherzo.keyboard_octave = 4;
    scherzo.keyboard_accidental = SCORE_PITCH_ACCIDENTAL_NATURAL;

    note_group_init (scherzo.ctx, &scherzo.notes_pressed);
    note_group_init (scherzo.ctx, &scherzo.notes_pedaled);

    scherzo.pedal_pressed = 0;

    mnemon_init (&scherzo.mnemon);
    /* XXX: Should create a default file if one cannot be loaded. */
    mnemon_load_category (&scherzo.mnemon, "scherzo-notes");

    scherzo.challenge.note = NULL;
/*
    _select_challenge (&scherzo);
*/

    err = snd_midi_event_new (MIDI_BUF_SIZE, &scherzo.snd_midi_event);
    if (err) {
	fprintf (stderr, "Out of memory.\n");
	return 1;
    }

#define MIDI_DEVICE "/dev/midi1"
    scherzo.midi_fd = open (MIDI_DEVICE, O_RDONLY);
    if (scherzo.midi_fd < 0) {
	printf ("failed to open " MIDI_DEVICE ". Midi input will not be available.\n");
    } else {
	GIOChannel *channel;

	channel = g_io_channel_unix_new (scherzo.midi_fd);
	g_io_channel_set_encoding (channel, NULL, NULL);
	g_io_add_watch (channel, G_IO_IN, on_midi_input, &scherzo);
    }

    scherzo.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    gtk_window_set_default_size (GTK_WINDOW (scherzo.window), 1000, 600);

    g_signal_connect (scherzo.window, "delete-event",
		      G_CALLBACK (on_delete_event_quit), NULL);

    drawing_area = gtk_drawing_area_new ();

    gtk_container_add (GTK_CONTAINER (scherzo.window), drawing_area);

    g_signal_connect (drawing_area, "expose-event",  
		      G_CALLBACK (on_expose_event_draw),
		      &scherzo);

    g_signal_connect (scherzo.window, "key-press-event",
		      G_CALLBACK (on_key_press_event),
		      &scherzo);

    g_signal_connect (scherzo.window, "key-release-event",
		      G_CALLBACK (on_key_release_event),
		      &scherzo);
    
    gtk_widget_show_all (scherzo.window);
    
    gtk_main ();

    mnemon_save (&scherzo.mnemon);

    snd_midi_event_free (scherzo.snd_midi_event);

    talloc_free (scherzo.ctx);

    return 0;
}
