/* scherzo - Music notation training
 *
 * 	score - Utilities for drawing (simple) musical scores
 *
 * Copyright © 2010 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 */

#ifndef SCORE_H
#define SCORE_H

#include <talloc.h>
#include <cairo.h>

typedef struct score score_t;
typedef struct score_staff score_staff_t;

#define SCORE_PITCH_ACCIDENTAL_MASK 0x07
#define SCORE_PITCH_ACCIDENTAL_SHIFT 0

typedef enum score_pitch_accidental
{
    SCORE_PITCH_ACCIDENTAL_DOUBLE_FLAT,
    SCORE_PITCH_ACCIDENTAL_FLAT,
    SCORE_PITCH_ACCIDENTAL_NATURAL,
    SCORE_PITCH_ACCIDENTAL_SHARP,
    SCORE_PITCH_ACCIDENTAL_DOUBLE_SHARP
} score_pitch_accidental_t;

#define SCORE_PITCH_ACCIDENTAL(pitch) (((pitch) & SCORE_PITCH_ACCIDENTAL_MASK) >> SCORE_PITCH_ACCIDENTAL_SHIFT)

#define SCORE_PITCH_NAME_MASK 0x38
#define SCORE_PITCH_NAME_SHIFT 3

typedef enum score_pitch_name
{
    SCORE_PITCH_NAME_C,
    SCORE_PITCH_NAME_D,
    SCORE_PITCH_NAME_E,
    SCORE_PITCH_NAME_F,
    SCORE_PITCH_NAME_G,
    SCORE_PITCH_NAME_A,
    SCORE_PITCH_NAME_B,
} score_pitch_name_t;

#define SCORE_PITCH_NAME(pitch) (((pitch) & SCORE_PITCH_NAME_MASK) >> SCORE_PITCH_NAME_SHIFT)

#define SCORE_PITCH(name, accidental) (((name) << SCORE_PITCH_NAME_SHIFT) | (accidental))

#define SCORE_PITCH_LITERAL(literal_name, literal_accidental) SCORE_PITCH(SCORE_PITCH_NAME_##literal_name, SCORE_PITCH_ACCIDENTAL_##literal_accidental)

typedef enum score_pitch
{
    SCORE_PITCH_Cff = SCORE_PITCH_LITERAL (C, DOUBLE_FLAT), 
    SCORE_PITCH_Cf  = SCORE_PITCH_LITERAL (C, FLAT),	  
    SCORE_PITCH_C   = SCORE_PITCH_LITERAL (C, NATURAL),	  
    SCORE_PITCH_Cs  = SCORE_PITCH_LITERAL (C, SHARP),	  
    SCORE_PITCH_Css = SCORE_PITCH_LITERAL (C, DOUBLE_SHARP),

    SCORE_PITCH_Dff = SCORE_PITCH_LITERAL (D, DOUBLE_FLAT), 
    SCORE_PITCH_Df  = SCORE_PITCH_LITERAL (D, FLAT),	  
    SCORE_PITCH_D   = SCORE_PITCH_LITERAL (D, NATURAL),	  
    SCORE_PITCH_Ds  = SCORE_PITCH_LITERAL (D, SHARP),	  
    SCORE_PITCH_Dss = SCORE_PITCH_LITERAL (D, DOUBLE_SHARP),

    SCORE_PITCH_Eff = SCORE_PITCH_LITERAL (E, DOUBLE_FLAT), 
    SCORE_PITCH_Ef  = SCORE_PITCH_LITERAL (E, FLAT),	  
    SCORE_PITCH_E   = SCORE_PITCH_LITERAL (E, NATURAL),	  
    SCORE_PITCH_Es  = SCORE_PITCH_LITERAL (E, SHARP),	  
    SCORE_PITCH_Ess = SCORE_PITCH_LITERAL (E, DOUBLE_SHARP),

    SCORE_PITCH_Fff = SCORE_PITCH_LITERAL (F, DOUBLE_FLAT), 
    SCORE_PITCH_Ff  = SCORE_PITCH_LITERAL (F, FLAT),	  
    SCORE_PITCH_F   = SCORE_PITCH_LITERAL (F, NATURAL),	  
    SCORE_PITCH_Fs  = SCORE_PITCH_LITERAL (F, SHARP),	  
    SCORE_PITCH_Fss = SCORE_PITCH_LITERAL (F, DOUBLE_SHARP),

    SCORE_PITCH_Gff = SCORE_PITCH_LITERAL (G, DOUBLE_FLAT), 
    SCORE_PITCH_Gf  = SCORE_PITCH_LITERAL (G, FLAT),	  
    SCORE_PITCH_G   = SCORE_PITCH_LITERAL (G, NATURAL),	  
    SCORE_PITCH_Gs  = SCORE_PITCH_LITERAL (G, SHARP),	  
    SCORE_PITCH_Gss = SCORE_PITCH_LITERAL (G, DOUBLE_SHARP),

    SCORE_PITCH_Aff = SCORE_PITCH_LITERAL (A, DOUBLE_FLAT), 
    SCORE_PITCH_Af  = SCORE_PITCH_LITERAL (A, FLAT),	  
    SCORE_PITCH_A   = SCORE_PITCH_LITERAL (A, NATURAL),	  
    SCORE_PITCH_As  = SCORE_PITCH_LITERAL (A, SHARP),	  
    SCORE_PITCH_Ass = SCORE_PITCH_LITERAL (A, DOUBLE_SHARP),

    SCORE_PITCH_Bff = SCORE_PITCH_LITERAL (B, DOUBLE_FLAT), 
    SCORE_PITCH_Bf  = SCORE_PITCH_LITERAL (B, FLAT),	  
    SCORE_PITCH_B   = SCORE_PITCH_LITERAL (B, NATURAL),	  
    SCORE_PITCH_Bs  = SCORE_PITCH_LITERAL (B, SHARP),	  
    SCORE_PITCH_Bss = SCORE_PITCH_LITERAL (B, DOUBLE_SHARP)
} score_pitch_t;

typedef enum score_duration
{
    SCORE_DURATION_WHOLE = 1,
    SCORE_DURATION_1 = 1,
    SCORE_DURATION_HALF = 2,
    SCORE_DURATION_2 = 2,
    SCORE_DURATION_QUARTER = 4,
    SCORE_DURATION_4 = 4,
    SCORE_DURATION_EIGHTH = 8,
    SCORE_DURATION_8 = 8,
    SCORE_DURATION_SIXTEENTH = 16,
    SCORE_DURATION_16 = 16,
    SCORE_DURATION_THIRTYSECOND = 32,
    SCORE_DURATION_32 = 32,
    SCORE_DURATION_SIXTYFOURTH = 64,
    SCORE_DURATION_64 = 64,
    SCORE_DURATION_ONEHUNDREDTWENTYEIGHTH = 128,
    SCORE_DURATION_128 = 128
} score_duration_t;

#define SCORE_BUILD_NOTE(pitch, octave, duration) SCORE_PITCH_##pitch, (octave), SCORE_DURATION_##duration

typedef struct score_note
{
    score_staff_t *staff;
    score_pitch_t pitch;
    int octave;
    score_duration_t duration;

    struct {
	double r;
	double g;
	double b;
    } color;
} score_note_t;

typedef struct score_chord
{
    score_staff_t *staff;

    char *name;
    double width;
} score_chord_t;

typedef enum score_clef
{
    SCORE_CLEF_G,
    SCORE_CLEF_TREBLE = SCORE_CLEF_G,
    SCORE_CLEF_F,
    SCORE_CLEF_BASS = SCORE_CLEF_F
} score_clef_t;

/* Allocate a new, empty score object, (with optional ctx as talloc
 * owner). If ctx is NULL, the caller should call talloc_free on the
 * score_t* when done with it. Otherwise, the object will be freed
 * when ctx is freed. */
score_t *
score_create (void *ctx);

/* Set an (approximate) staff height. The actual staff height may
 * differ due to rounding to achieve evenly spaced, sharply rendered
 * lines. the actual staff height is returned. */
int
score_set_staff_height (score_t *score, int height);

/* Set the total width available for drawing the score. */
void
score_set_width (score_t *score, int width);

/* Add a brace to the score, connecting the given number of staves.
 *
 * The staves to be connected are those that will next be added to the
 * score. */
void
score_add_brace (score_t *score, int staves);

/* Add a new staff to the score */
score_staff_t *
score_add_staff (score_t *score, score_clef_t clef);

/* Add a note to a staff of the given pitch, octave, and duration.
 *
 * Octave numbers are ISO octave numbers [0:8], (so Octave 4 is from
 * middle C to the B above middle C).
 *
 * Duration values can be symbolic (SCORE_DURATION_WHOLE, _QUARTER,
 * _EIGHTH, etc.) or numerical as simply the denominator (WHOLE=1,
 * QUARTER=4, EIGHTH=8, etc.)
 */
score_note_t *
score_add_note (score_staff_t *staff,
		score_pitch_t pitch,
		int octave,
		score_duration_t);

/* Add a chord symbol of 'name' to a staff.
 *
 * For now, the chord symbols are free-form names.
 *
 * The chord name must be a talloc'ed string, which the returned
 * score_chord_t will talloc_steal.
 */
score_chord_t *
score_add_chord (score_staff_t *staff,
		 const char * name);

/* Remove the given chord from its staff. */
void
score_remove_chord (score_chord_t *chord);

/* Remove the given note from its staff. */
void
score_remove_note (score_note_t *note);

void
score_set_note_color_rgb (score_note_t *note,
			  double r,
			  double g,
			  double b);

/* Return the first note on the given staff with the given pitch,
 * octave, and durations. Returns NULL if no match is found. */
score_note_t *
score_staff_find_note (score_staff_t *staff,
		       score_pitch_t pitch,
		       int octave,
		       score_duration_t duration);

/* Draw the given score_t onto the given cairo_t.
 *
 * The caller can call cairo_translate before calling score_draw to
 * position the result as desired, (and can call cairo_clip to clip it
 * if desired). */
void
score_draw (score_t *score, cairo_t *cr);

#endif
